# uci.py
# Copyright 2015 Roger Marsh
# License: See LICENSE.TXT (BSD licence)

"""Control multiple UCI compliant chess engines."""

import tkinter
import tkinter.messagebox
import tkinter.filedialog

from solentware_grid.core.dataclient import DataSource

from solentware_misc.gui.exceptionhandler import ExceptionHandler

from uci_net.engine import (
    ReservedOptionNames,
)

from ..core.uci import UCI as _UCI
from .enginegrid import EngineGrid
from .enginerow import make_ChessDBrowEngine
from ..core.filespec import (
    ENGINE_FILE_DEF,
    COMMAND_FIELD_DEF,
)
from .eventspec import EventSpec

import sys

_win32_platform = sys.platform == "win32"
_freebsd_platform = sys.platform.startswith("freebsd")
del sys


class UCI(ExceptionHandler):
    def __init__(self, menu_engines, menu_commands):

        self._do_toplevel = None
        self._show_engines_toplevel = None
        self.database = None
        self.base_engines = None
        self.visible_scrollbars = True

        # The chess engine definition widgets in their own Toplevel.
        # Should these be helper class responsibility as well?
        # Similar construct in .chess_ui.ChessUI too.
        self.engines_in_toplevels = set()

        self.menu_engines = menu_engines
        self.menu_commands = menu_commands

        menu_engines.add_separator()
        for accelerator, function in (
            (
                EventSpec.menu_engines_position_queues,
                self.position_queue_status,
            ),
            (EventSpec.menu_engines_show_engines, self.show_engines),
            (EventSpec.menu_engines_start_engine, self.start_engine),
            (EventSpec.menu_engines_quit_all_engines, self.quit_all_engines),
        ):
            menu_engines.add_command(
                label=accelerator[1],
                command=self.try_command(function, menu_engines),
                underline=accelerator[3],
            )
        menu_engines.add_separator()
        menu_engines.add_separator()

        menu_commands.add_separator()
        for accelerator, function in (
            (EventSpec.menu_commands_multipv, self.set_multipv),
            (EventSpec.menu_commands_depth, self.set_depth),
            (EventSpec.menu_commands_hash, self.set_hash),
        ):
            menu_commands.add_command(
                label=accelerator[1],
                command=self.try_command(function, menu_commands),
                underline=accelerator[3],
            )
        menu_commands.add_separator()

        # Must do both these commands between each move to get consistent
        # analysis between runs of ChessTab given that positions are evaluated
        # in random order.
        # menu_commands.add_command(
        #    label='Ucinewgame off',
        #    underline=0,
        #    command=self.try_command(self.set_ucinewgame_off, menu_commands))
        # menu_commands.add_command(
        #    label='Clear Hash off',
        #    underline=0,
        #    command=self.try_command(
        #        self.set_clear_hash_off, menu_commands))
        # menu_commands.add_separator()

        # Not implemented at present.
        # And possibly never will be.
        # menu_commands.add_command(
        #    label='Go infinite',
        #    underline=3,
        #    command=self.try_command(self.go_infinite, menu_commands))
        # menu_commands.add_command(
        #    label='Stop',
        #    underline=0,
        #    command=self.try_command(self.stop, menu_commands))
        # menu_commands.add_separator()
        # menu_commands.add_command(
        #    label='Isready',
        #    underline=2,
        #    command=self.try_command(self.isready, menu_commands))
        # menu_commands.add_separator()

        self.uci = _UCI()

    @property
    def show_engines_toplevel(self):
        """ """
        return self._show_engines_toplevel

    def remove_engines_and_menu_entries(self):
        """Quit all started UCI compliant Chess Engines."""
        ui_names = set(self.uci.uci_drivers)
        still_alive = self.uci.quit_all_engines()
        ui_names -= set([n for n, p in still_alive])
        dead_menu_items = []
        for i in range(self.menu_engines.index("end") + 1):
            label = self.menu_engines.entryconfigure(i).get("label")
            if label is not None:
                if label[-1] in ui_names:
                    dead_menu_items.insert(0, i)
        for dmi in dead_menu_items:
            self.menu_engines.delete(dmi)
        for name, pid in still_alive:
            tkinter.messagebox.showinfo(
                parent=self.menu_commands,
                title="Stop Engine",
                message="".join(
                    (
                        name,
                        " failed to stop.\n\nYou may have to kill",
                        " process id ",
                        str(pid),
                        " manually.",
                    )
                ),
            )

    def quit_all_engines(self):
        """Quit all started UCI compliant Chess Engines after confirmation."""
        if (
            tkinter.messagebox.askquestion(
                parent=self.menu_commands,
                title="Engines",
                message="Confirm Quit All Engines.",
            )
            == tkinter.messagebox.YES
        ):
            self.remove_engines_and_menu_entries()

    def start_engine(self):
        """Start an UCI compliant Chess Engine."""

        # Avoid "OSError: [WinError 535] Pipe connected"  at Python3.3 running
        # under Wine on FreeBSD 10.1 by disabling the UCI functions.
        # Assume all later Pythons are affected because they do not install
        # under Wine at time of writing.
        # The OSError stopped happening by wine-2.0_3,1 on FreeBSD 10.1 but
        # get_nowait() fails to 'not wait', so ChessTab never gets going under
        # wine at present.  Leave alone because it looks like the problem is
        # being shifted constructively.
        # At Python3.5 running under Wine on FreeBSD 10.1, get() does not wait
        # when the queue is empty either, and ChessTab does not run under
        # Python3.3 because it uses asyncio: so no point in disabling.
        # if self.uci.uci_drivers_reply is None:
        #    tkinter.messagebox.showinfo(
        #        parent=self.menu_commands,
        #        title='Chesstab Restriction',
        #        message=' '.join(
        #            ('Starting an UCI chess engine is not allowed because',
        #             'an interface cannot be created:',
        #             'this is expected if running under Wine.')))
        #    return

        if _win32_platform:
            filetypes = (("Chess Engines", "*.exe"),)
        else:
            filetypes = ()
        filename = tkinter.filedialog.askopenfilename(
            parent=self.menu_commands,
            title="Run Chess Engine",
            filetypes=filetypes,
            initialfile="",
            initialdir="~",
        )
        if not filename:
            return

        def get(event):
            command = self._contents.get()
            if command == filename:
                self.run_engine(filename)
            elif not command.startswith(filename):
                tkinter.messagebox.showinfo(
                    parent=self._do_toplevel,
                    title="Start Engine",
                    message="Command must start with selected file name.",
                )
            else:
                command = command.replace(filename, "", 1)
                if not command.startswith(" "):
                    tkinter.messagebox.showinfo(
                        parent=self._do_toplevel,
                        title="Start Engine",
                        message="Command must start with selected file name.",
                    )
                else:
                    self.run_engine(filename, args=command.strip())
            self._cancel()

        self.do_command(
            filename,
            get,
            hint="".join(
                (
                    "\nConsult chess engine's documentation for any arguments ",
                    "which must be appended to the command.\n\nPress enter to ",
                    "run engine.\n",
                )
            ),
        )

    def _cancel(self):
        self._do_toplevel.destroy()
        self._do_toplevel = None
        del self._contents

    def do_command(self, initial_value, callback, hint=None, wraplength=None):

        if self._do_toplevel is not None:
            tkinter.messagebox.showinfo(
                parent=self.menu_commands,
                title="Engine Command",
                message="".join(
                    (
                        "A command dialogue is already active:\n\n",
                        "Cannot start another one.",
                    )
                ),
            )
            return
        self._command = initial_value.split(None, maxsplit=1)[0]
        self._do_toplevel = tkinter.Toplevel(
            master=self.menu_engines.winfo_toplevel()
        )
        if hint:
            if wraplength is None:
                wraplength = 400
            tkinter.Label(
                master=self._do_toplevel,
                text=hint,
                wraplength=wraplength,
                justify=tkinter.LEFT,
            ).pack()
        entrythingy = tkinter.Entry(self._do_toplevel)
        entrythingy.pack(fill=tkinter.BOTH)
        buttonbar = tkinter.Frame(master=self._do_toplevel)
        buttonbar.pack(fill=tkinter.X, expand=tkinter.TRUE)
        cancel = tkinter.Button(
            master=buttonbar,
            text="Cancel",
            underline=0,
            command=self.try_command(self._cancel, buttonbar),
        )
        cancel.pack(expand=tkinter.TRUE, side=tkinter.LEFT)
        self._contents = tkinter.StringVar()
        self._contents.set(initial_value)
        entrythingy["textvariable"] = self._contents
        entrythingy.bind("<Key-Return>", callback)
        entrythingy.focus_set()

    def run_engine(self, program_file_name, args=None):

        self.uci.run_engine(program_file_name, args)
        self.menu_engines.insert_command(
            self.menu_engines.index("end"),
            label=self.uci.uci_drivers_index[self.uci.engine_counter],
            command=self.make_quit_engine_command(self.uci.engine_counter),
        )

    def make_quit_engine_command(self, counter):
        def quit_engine_command():
            self.quit_engine(counter)

        return quit_engine_command

    def quit_engine(self, number):

        ui_name = self.uci.uci_drivers_index[number]
        for i in range(self.menu_engines.index("end")):
            label = self.menu_engines.entryconfigure(i).get("label")
            if label is not None:
                if label[-1] == ui_name:
                    if (
                        tkinter.messagebox.askquestion(
                            parent=self.menu_engines,
                            title="Quit Engine",
                            message="".join(
                                (
                                    "Please confirm request to quit engine\n\n",
                                    ui_name,
                                )
                            ),
                        )
                        != tkinter.messagebox.YES
                    ):
                        continue
                    if self.uci.kill_engine(number):
                        self.menu_engines.delete(i)
                    else:
                        tkinter.messagebox.showinfo(
                            parent=self.menu_engines,
                            title="Quit Engine",
                            message="".join(
                                (
                                    ui_name,
                                    " failed to quit.\n\nYou may have to kill",
                                    " process id ",
                                    str(
                                        self.uci.uci_drivers[
                                            ui_name
                                        ].driver.pid
                                    ),
                                    " manually.",
                                )
                            ),
                        )

    def send_to_all_engines(self, event):

        command = self._contents.get()
        if command.split()[0] != self._command:
            if (
                tkinter.messagebox.askquestion(
                    parent=self.menu_commands,
                    title="Send to Engine",
                    message="".join(
                        (
                            "Command is not the one used to start dialogue.\n\n",
                            "Do you want to cancel dialogue?",
                        )
                    ),
                )
                == tkinter.messagebox.YES
            ):
                del self._command
                del self._contents
                self._do_toplevel.destroy()
                self._do_toplevel = None
            return
        for ui_name, ei in self.uci.uci_drivers.items():
            try:
                ei.to_driver_queue.put(command)
            except:
                tkinter.messagebox.showinfo(
                    parent=self.menu_commands,
                    title="Send to Engine",
                    message="".join(
                        (
                            "Send command\n\n",
                            command,
                            "\n\nto\n\n",
                            ui_name,
                            "\n\nfailed.",
                        )
                    ),
                )
        del self._command
        del self._contents
        self._do_toplevel.destroy()
        self._do_toplevel = None

    def set_hash(self):
        """Set value of Hash option to use on UCI compliant Chess Engine."""

        def get(event):
            self.uci.hash_size = self._contents.get()
            self.uci.set_option_on_empty_queues.add(ReservedOptionNames.Hash)
            self._do_toplevel.destroy()
            self._do_toplevel = None
            del self._spinbox

        self.do_spinbox(
            self.uci.hash_size,
            0,
            1000,
            get,
            hint="".join(
                (
                    "\nSet the size of hash tables, in mega-bytes, within ",
                    "limits allowed by chess engines.\n\n0 (zero) means use ",
                    "the default assumed by a chess engine.\n\nPress enter to ",
                    "set value.\n",
                )
            ),
        )

    def set_multipv(self):
        """Set value of MultiPV option to use on UCI compliant Chess Engine."""

        def get(event):
            self.uci.multipv = self._contents.get()
            self._do_toplevel.destroy()
            self._do_toplevel = None
            del self._spinbox

        self.do_spinbox(
            self.uci.multipv,
            0,
            100,
            get,
            hint="".join(
                (
                    "\nSet the number of variations returned by chess engines ",
                    "which support the MultiPV option.\n\n0 (zero) means use ",
                    "the default assumed by a chess engine.\n\nPress enter to ",
                    "set value.\n",
                )
            ),
        )

    def set_depth(self):
        """Set depth to use in go commands to UCI compliant Chess Engine."""

        def get(event):
            self.uci.go_depth = self._contents.get()
            self._do_toplevel.destroy()
            self._do_toplevel = None
            del self._spinbox

        self.do_spinbox(
            self.uci.go_depth,
            0,
            200,
            get,
            hint="".join(
                (
                    "\nSet the length of variations, in half-moves, returned ",
                    "by chess engines.\n\nPress enter to set value.\n",
                )
            ),
        )

    def _modify_command_menu_item(self, old, new, command):
        for i in range(self.menu_commands.index("end")):
            label = self.menu_commands.entryconfigure(i).get("label")
            if label is not None:
                if label[-1] == old:
                    self.menu_commands.entryconfigure(
                        i,
                        label=new,
                        command=self.try_command(command, self.menu_commands),
                    )

    def set_ucinewgame_off(self):
        """Set to not use ucinewgame command when navigating games."""
        if (
            tkinter.messagebox.askquestion(
                parent=self.menu_commands,
                title="Ucinewgame OFF",
                message="".join(
                    (
                        "Turn use of ucinewgame command OFF when navigating in or ",
                        "between game scores?\n\n",
                        "Consult engine documentation for implications of using, ",
                        "or not using, the ucinewgame command.\n\n",
                        "UCI specification states new GUIs should support ucinewgame ",
                        "command, and engines should not expect ucinewgame commands ",
                        "if the ucinewgame command is not used before the first ",
                        "position command. (August 2015)",
                    )
                ),
            )
            == tkinter.messagebox.YES
        ):
            self.uci.use_ucinewgame = False
            self._modify_command_menu_item(
                "Ucinewgame off", "Ucinewgame on", self.set_ucinewgame_on
            )

    def set_ucinewgame_on(self):
        """Set to use ucinewgame command when navigating games."""
        if (
            tkinter.messagebox.askquestion(
                parent=self.menu_commands,
                title="Ucinewgame ON",
                message="".join(
                    (
                        "Turn use of ucinewgame command ON when navigating in or ",
                        "between game scores?\n\n",
                        "Consult engine documentation for implications of using, ",
                        "or not using, the ucinewgame command.\n\n",
                        "UCI specification states new GUIs should support ucinewgame ",
                        "command, and engines should not expect ucinewgame commands ",
                        "if the ucinewgame command is not used before the first ",
                        "position command. (August 2015)",
                    )
                ),
            )
            == tkinter.messagebox.YES
        ):
            self.uci.use_ucinewgame = True
            self._modify_command_menu_item(
                "Ucinewgame on", "Ucinewgame off", self.set_ucinewgame_off
            )

    def stop(self):
        """Send stop command to UCI compliant Chess Engines."""

        tkinter.messagebox.showinfo(
            parent=self.menu_commands,
            title="Stop Command",
            message="Stop not implemented.",
        )

    def go_infinite(self):
        """Send go infinite command to UCI compliant Chess Engines."""

        tkinter.messagebox.showinfo(
            parent=self.menu_commands,
            title="Go Infinite Command",
            message="Go infinite not implemented.",
        )

    def isready(self):
        """Send isready command to UCI compliant Chess Engines."""

        tkinter.messagebox.showinfo(
            parent=self.menu_commands,
            title="Isready Command",
            message="Isready not implemented.",
        )

    def do_spinbox(
        self, initial_value, from_, to, callback, hint=None, wraplength=None
    ):

        # Comapare with show_engines() method.
        # The after() technique used in show_engines does not help here to
        # guarantee display of the new Toplevel on second and subsequent use.
        # But the problem is seen only when the X server is on a different box
        # than the ChessTab process.

        if self._do_toplevel is not None:
            tkinter.messagebox.showinfo(
                parent=self.menu_commands,
                title="Engine Command",
                message="".join(
                    (
                        "A command dialogue is already active:\n\n",
                        "Cannot start another one.",
                    )
                ),
            )
            return

        def cancel():
            self._do_toplevel.destroy()
            self._do_toplevel = None
            del self._spinbox

        def destroy(event=None):
            self._do_toplevel = None

        self._spinbox = initial_value
        self._do_toplevel = tkinter.Toplevel(
            master=self.menu_engines.winfo_toplevel()
        )
        self._do_toplevel.bind("<Destroy>", destroy)
        if hint:
            if wraplength is None:
                wraplength = 400
            tkinter.Label(
                master=self._do_toplevel,
                text=hint,
                wraplength=wraplength,
                justify=tkinter.LEFT,
            ).pack()
        entrythingy = tkinter.Spinbox(self._do_toplevel)
        entrythingy.pack(fill=tkinter.BOTH)
        buttonbar = tkinter.Frame(master=self._do_toplevel)
        buttonbar.pack(fill=tkinter.X, expand=tkinter.TRUE)
        tkinter.Button(
            master=buttonbar,
            text="Cancel",
            underline=0,
            command=self.try_command(cancel, buttonbar),
        ).pack(expand=tkinter.TRUE, side=tkinter.LEFT)
        entrythingy.configure(from_=from_, to=to, increment=1)
        self._contents = tkinter.IntVar()
        self._contents.set(initial_value)
        entrythingy["textvariable"] = self._contents
        entrythingy.bind("<Key-Return>", callback)
        entrythingy.focus_set()

    def show_engines(self):
        """Show Chess Engine Descriptions on database."""
        if self._show_engines_toplevel is not None:
            tkinter.messagebox.showinfo(
                parent=self.menu_engines,
                title="Show Engines",
                message="".join(
                    (
                        "A show engines dialogue is already active:\n\n",
                        "Cannot start another one.",
                    )
                ),
            )
            return

        def destroy(event=None):
            if self._show_engines_toplevel is not None:
                self._close_engine_grid()
            self._show_engines_toplevel = None

        self._show_engines_toplevel = tkinter.Toplevel(
            master=self.menu_engines.winfo_toplevel()
        )
        self._show_engines_toplevel.wm_title("Chess Engines")
        self._show_engines_toplevel.bind("<Destroy>", destroy)
        self._close_engine_grid()

        # The after_idle version makes it less common for the Toplevel to not
        # become visible until 'click on menubar' or 'move pointer out of
        # application window' on 'destroy or close' toplevel and 'show engine'
        # action sequences.
        # The after version gives a noticably slower response when the X server
        # is on a different machine, but has not yet needed the pointer prompt
        # to display the new toplevel.
        # The problem has not been seen yet on Microsoft Windows XP.
        # The problem happens rarely when the X server is on the same machine.
        # The problem had never been seen before in the toplevels started from
        # the other subclasses of DataGrid: but on looking it can be seen now.
        # self._open_engine_grid()
        # self._show_engines_toplevel.after_idle(self.try_command(
        #    self._open_engine_grid, self._show_engines_toplevel))
        self._show_engines_toplevel.after(
            1,
            self.try_command(
                self._open_engine_grid, self._show_engines_toplevel
            ),
        )

    def set_open_database_and_engine_classes(self, database=None):
        """Set current open database and engine specific dataset classes."""
        if self.database is not database:
            self._close_engine_grid()
            self.database = database
            self._open_engine_grid()

    def _close_engine_grid(self):
        """Close the existing EngineGrid instance."""
        if self.base_engines is not None:
            self.base_engines.get_top_widget().pack_forget()
            self.base_engines.set_data_source()
            self.base_engines = None

    def _open_engine_grid(self):
        """Open and show an EngineGrid instance."""
        if self._show_engines_toplevel is None:
            return
        if self.base_engines is None and self.database is not None:
            self.base_engines = EngineGrid(self)
            self.base_engines.set_data_source(
                DataSource(
                    self.database,
                    ENGINE_FILE_DEF,
                    COMMAND_FIELD_DEF,
                    make_ChessDBrowEngine(self),
                ),
                self.base_engines.on_data_change,
            )
            self.base_engines.set_focus()
            self.base_engines.get_top_widget().pack(
                fill=tkinter.BOTH, expand=tkinter.TRUE
            )

    def show_scrollbars(self):
        """Show the scrollbars in the engine definition display widgets."""
        self.visible_scrollbars = True
        exceptions = []
        for items in (self.engines_in_toplevels,):
            for i in items:
                try:
                    i.show_scrollbars()
                except tkinter.TclError:
                    exceptions.append(i, items)
        for i, items in exceptions:
            items.remove(i)

    def hide_scrollbars(self):
        """Show the scrollbars in the engine definition display widgets."""
        self.visible_scrollbars = False
        exceptions = []
        for items in (self.engines_in_toplevels,):
            for i in items:
                try:
                    i.hide_scrollbars()
                except tkinter.TclError:
                    exceptions.append((i, items))
        for i, items in exceptions:
            items.remove(i)

    def position_queue_status(self):
        """Display counts of position queued for analysis by engines."""
        pending_counts = []
        for engine, pending in self.uci.positions_pending.items():
            pending_counts.append(
                (engine, sum([len(p) for p in pending.values()]))
            )
        if not pending_counts:
            tkinter.messagebox.showinfo(
                parent=self.menu_commands,
                title="Position Queues",
                message="There are no queues of positions for analysis.",
            )
            return
        pce = []
        for e, pc in sorted(pending_counts):
            pce.append("\t".join((str(pc), e)))
        tkinter.messagebox.showinfo(
            parent=self.menu_commands,
            title="Position Queues",
            message="".join(
                (
                    "Number of positions queued for analysis by ",
                    "each engine are:\n\n",
                    "\n".join(pce),
                )
            ),
        )

    def set_clear_hash_off(self):
        """Set to not clear hash tables before position go command sequence."""
        if (
            tkinter.messagebox.askquestion(
                parent=self.menu_commands,
                title="Clear Hash OFF",
                message="".join(
                    (
                        "Turn OFF clear hash tables before analysing a position?\n\n",
                        "Consult engine documentation for implications of clearing ",
                        "hash tables or not.",
                    )
                ),
            )
            == tkinter.messagebox.YES
        ):
            self.uci.clear_hash = False
            self._modify_command_menu_item(
                "Clear Hash off", "Clear Hash on", self.set_clear_hash_on
            )

    def set_clear_hash_on(self):
        """Set to clear hash tables before position go command sequence."""
        if (
            tkinter.messagebox.askquestion(
                parent=self.menu_commands,
                title="Clear Hash ON",
                message="".join(
                    (
                        "Turn ON clear hash tables before analysing a position?\n\n",
                        "Consult engine documentation for implications of clearing ",
                        "hash tables or not.",
                    )
                ),
            )
            == tkinter.messagebox.YES
        ):
            self.uci.clear_hash = True
            self._modify_command_menu_item(
                "Clear Hash on", "Clear Hash off", self.set_clear_hash_off
            )



# queryrow.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Create widgets that display details of game selection rules on database.
"""

import tkinter

from solentware_grid.gui.datarow import (
    GRID_COLUMNCONFIGURE,
    GRID_CONFIGURE,
    WIDGET_CONFIGURE,
    WIDGET,
    ROW,
)

from .datarow import DataRow
from ..core.chessrecord import ChessDBrecordQuery
from .querydbedit import QueryDbEdit
from .querydbdelete import QueryDbDelete
from .querydbshow import QueryDbShow
from . import constants

ON_DISPLAY_COLOUR = "#eba610"  # a pale orange


class ChessDBrowQuery(ChessDBrecordQuery, DataRow):

    """Define row in list of game selection rules.

    Add row methods to the game selection rule record definition.

    """

    header_specification = [
        {
            WIDGET: tkinter.Label,
            WIDGET_CONFIGURE: dict(
                text="Description",
                anchor=tkinter.W,
                padx=0,
                pady=1,
                font="TkDefaultFont",
            ),
            GRID_CONFIGURE: dict(column=0, sticky=tkinter.EW),
            GRID_COLUMNCONFIGURE: dict(weight=1, uniform="pp"),
            ROW: 0,
        },
    ]

    def __init__(self, database=None, ui=None):
        """Extend and associate record definition with database.

        database - the open database that is source of row data
        ui - the ChessUI instamce

        """
        super().__init__()
        self.ui = ui
        self.set_database(database)
        self.row_specification = [
            {
                WIDGET: tkinter.Label,
                WIDGET_CONFIGURE: dict(
                    anchor=tkinter.W,
                    font=constants.LISTS_OF_GAMES_FONT,
                    pady=1,
                    padx=0,
                ),
                GRID_CONFIGURE: dict(column=0, sticky=tkinter.EW),
                ROW: 0,
            },
        ]

    def show_row(self, dialog, oldobject):
        """Return a QueryDbShow toplevel for oldobject.

        dialog - a Toplevel
        oldobject - a ChessDBrecordQuery containing original data

        """
        return QueryDbShow(dialog, oldobject, ui=self.ui)

    def delete_row(self, dialog, oldobject):
        """Return a QueryDbDelete dialog for oldobject.

        dialog - a Toplevel
        oldobject - a ChessDBrecordQuery containing original data

        """
        return QueryDbDelete(dialog, oldobject, ui=self.ui)

    def edit_row(self, dialog, newobject, oldobject, showinitial=True):
        """Return a QueryDbEdit dialog for oldobject.

        dialog - a Toplevel
        newobject - a ChessDBrecordQuery containing original data to be
                    edited
        oldobject - a ChessDBrecordQuery containing original data
        showintial == True - show both original and edited data

        """
        return QueryDbEdit(
            newobject, dialog, oldobject, showinitial=showinitial, ui=self.ui
        )

    def grid_row(self, **kargs):
        """Return super().grid_row(textitems=(...), **kargs).

        Create textitems argument for ChessDBrowQuery instance.

        """
        return super().grid_row(
            textitems=(
                self.value.get_name_text(),
                # self.value.get_selection_rule_text(),
            ),
            **kargs
        )

    def grid_row_on_display(self, **kargs):
        self._current_row_background = ON_DISPLAY_COLOUR
        return self.grid_row(background=ON_DISPLAY_COLOUR, **kargs)

    def set_background_on_display(self, widgets):
        self._current_row_background = ON_DISPLAY_COLOUR
        self.set_background(widgets, self._current_row_background)


def make_ChessDBrowQuery(chessui):
    """Make ChessDBrowQuery with reference to ChessUI instance"""

    def make_selection(database=None):
        return ChessDBrowQuery(database=database, ui=chessui)

    return make_selection



# __init__.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Access a Berkeley DB database of chess games using the bsddb3 package.

This interface can be used if the bsddb3 package is installed.

The bsddb3 package is available as a source package, from PyPI for example.

Follow the instructions to build and install bsddb3.  This includes installing
Berkeley DB.
"""



# queryedit.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Edit a game selection rule and change main list of games to fit.

The QueryEdit class extends the query.Query class to allow editing.

An instance of these classes fits into the user interface in two ways: as an
item in a panedwindow of the main widget, or as the only item in a new toplevel
widget.

"""

from .query import Query
from .eventspec import EventSpec


class QueryEdit(Query):

    """Display a game selection rule with editing allowed.

    Attribute _is_text_editable is True means the statement can be edited.

    """

    # True means selection selection can be edited
    _is_text_editable = True

    # Remove if this is all it is left doing.
    def __init__(self, **ka):
        """Extend game selection rule widget as editor."""
        super().__init__(**ka)



# chessdbdu.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database update using custom deferred update for Berkeley DB.
"""

import os
import bz2
import zipfile

from bsddb3.db import (
    DB_CREATE,
    DB_RECOVER,
    DB_INIT_MPOOL,
    DB_INIT_LOCK,
    DB_INIT_LOG,
    DB_INIT_TXN,
    DB_PRIVATE,
)

from solentware_base import bsddb3du_database

# from solentware_base.core.constants import FILEDESC
from solentware_base.core.constants import (
    FILEDESC,
    SUBFILE_DELIMITER,
    EXISTENCE_BITMAP_SUFFIX,
    SEGMENT_SUFFIX,
)
from solentware_base.core.segmentsize import SegmentSize

from ..core.filespec import (
    FileSpec,
    GAMES_FILE_DEF,
    DB_ENVIRONMENT_GIGABYTES,
    DB_ENVIRONMENT_BYTES,
    DB_ENVIRONMENT_MAXLOCKS,
    SECONDARY,
)
from ..core.chessrecord import ChessDBrecordGameImport


class Chessbsddb3duError(Exception):
    pass


def chess_dbdu(
    dbpath, pgnpaths, file_records, reporter=lambda text, timestamp=True: None
):
    """Open database, import games and close database."""
    cdb = ChessDatabase(dbpath, allowcreate=True)
    importer = ChessDBrecordGameImport()
    cdb.open_database()
    cdb.set_defer_update()
    for pp in pgnpaths:
        s = open(pp, "r", encoding="iso-8859-1")
        importer.import_pgn(cdb, s, pp, reporter=reporter)
        s.close()
    if reporter is not None:
        reporter("Finishing import: please wait.")
        reporter("", timestamp=False)
    cdb.do_final_segment_deferred_updates()
    cdb.unset_defer_update()
    cdb.close_database()

    # There are no recoverable file full conditions for Berkeley DB (see DPT).
    return True


class ChessDatabase(bsddb3du_database.Database):

    """Provide custom deferred update for a database of games of chess."""

    # The optimum chunk size is the segment size.
    # Assuming 2Gb memory:
    # A default FreeBSD user process will not cause a MemoryError exception for
    # segment sizes up to 65536 records, so the optimum chunk size defined in
    # the superclass will be selected.
    # A MS Windows XP process will cause the MemoryError exeption which selects
    # the 32768 game chunk size.
    # A default OpenBSD user process will cause the MemoryError exception which
    # selects the 16384 game chunk size.
    # The error log problem fixed at chesstab-0.41.9 obscured what was actually
    # happening: OpenBSD gives a MemoryError exception but MS Windows XP heads
    # for thrashing swap space in some runs with a 65536 chunk size (depending
    # on order of processing indexes I think). Windows 10 Task Manager display
    # made this obvious.
    # The MemoryError exception or swap space thrashing will likely not occur
    # for a default OpenBSD user process or a MS Windows XP process with segment
    # sizes up to 32768 records. Monitoring with Top and Task Manager suggests
    # it gets close with OpenBSD.
    if SegmentSize.db_segment_size > 32768:
        for f, m in ((4, 700000000), (2, 1400000000)):
            try:
                b" " * m
            except MemoryError:

                # Override the value in the superclass.
                deferred_update_points = frozenset(
                    [
                        i
                        for i in range(
                            65536 // f - 1,
                            SegmentSize.db_segment_size,
                            65536 // f,
                        )
                    ]
                )

                break
        del f, m

    def __init__(self, DBfile, **kargs):
        """Define chess database.

        **kargs
        allowcreate == False - remove file descriptions from FileSpec so
        that superclass cannot create them.
        Other arguments are passed through to superclass __init__.

        """
        names = FileSpec(**kargs)
        environment = {
            "flags": (
                DB_CREATE
                | DB_RECOVER
                | DB_INIT_MPOOL
                | DB_INIT_LOCK
                | DB_INIT_LOG
                | DB_INIT_TXN
                | DB_PRIVATE
            ),
            "gbytes": DB_ENVIRONMENT_GIGABYTES,
            "bytes": DB_ENVIRONMENT_BYTES,
            "maxlocks": DB_ENVIRONMENT_MAXLOCKS,
        }
        # Deferred update for games file only
        # for t in list(names.keys()):
        #    if t != GAMES_FILE_DEF:
        #        del names[t]

        if not kargs.get("allowcreate", False):
            try:
                for t in names:
                    if FILEDESC in names[t]:
                        del names[t][FILEDESC]
            except:
                if __name__ == "__main__":
                    raise
                else:
                    raise Chessbsddb3duError("DB description invalid")

        try:
            super(ChessDatabase, self).__init__(names, DBfile, environment)
        except Chessbsddb3duError:
            if __name__ == "__main__":
                raise
            else:
                raise Chessbsddb3duError("DB description invalid")

    def open_context_prepare_import(self):
        """Return True

        No preparation actions thet need database open for Berkeley DB.

        """
        return True

    def get_archive_names(self, files=()):
        """Return specified files and existing operating system files"""
        if self.home_directory is None:
            return None, []
        if self._file_per_database:
            names = dict()
            for k in self.specification:
                if k not in files:
                    continue
                ns = []
                names[os.path.join(self.home_directory, k)] = ns
                for i in self.specification[k][SECONDARY]:
                    ns.append(
                        os.path.join(
                            self.home_directory, SUBFILE_DELIMITER.join((k, i))
                        )
                    )
                ns.append(
                    os.path.join(
                        self.home_directory,
                        SUBFILE_DELIMITER.join((k, EXISTENCE_BITMAP_SUFFIX)),
                    )
                )
                ns.append(
                    os.path.join(
                        self.home_directory,
                        SUBFILE_DELIMITER.join((k, SEGMENT_SUFFIX)),
                    )
                )
            exists = [
                os.path.basename(n)
                for n in names
                if os.path.exists(".".join((n, "zip")))
            ]
            return names, exists
        else:
            names = [self.database_file]
            exists = [
                os.path.basename(n)
                for n in names
                if os.path.exists(".".join((n, "bz2")))
            ]
            return names, exists

    def archive(self, flag=None, names=None):
        """Write a bz2 or zip backup of files containing games.

        Intended to be a backup in case import fails.

        """
        if self.home_directory is None:
            return
        if names is None:
            return False
        if not self.delete_archive(flag=flag, names=names):
            return
        if flag:
            if self._file_per_database:
                for n in names:
                    archiveguard = ".".join((n, "grd"))
                    archivename = ".".join((n, "zip"))
                    c = zipfile.ZipFile(
                        archivename,
                        mode="w",
                        compression=zipfile.ZIP_DEFLATED,
                        allowZip64=True,
                    )
                    for s in names[n]:
                        c.write(s, arcname=os.path.basename(s))
                    c.close()
                    c = open(archiveguard, "wb")
                    c.close()
            else:
                for n in names:
                    c = bz2.BZ2Compressor()
                    archiveguard = ".".join((n, "grd"))
                    archivename = ".".join((n, "bz2"))
                    fi = open(n, "rb")
                    fo = open(archivename, "wb")
                    inp = fi.read(10000000)
                    while inp:
                        co = c.compress(inp)
                        if co:
                            fo.write(co)
                        inp = fi.read(10000000)
                    co = c.flush()
                    if co:
                        fo.write(co)
                    fo.close()
                    fi.close()
                    c = open(archiveguard, "wb")
                    c.close()
        return True

    def delete_archive(self, flag=None, names=None):
        """Delete a zip backup of files containing games."""
        if self.home_directory is None:
            return
        if names is None:
            return False
        if flag:
            if self._file_per_database:
                not_backups = []
                for n in names:
                    archiveguard = ".".join((n, "grd"))
                    archivename = ".".join((n, "zip"))
                    if not os.path.exists(archivename):
                        try:
                            os.remove(archiveguard)
                        except:
                            pass
                        continue
                    c = zipfile.ZipFile(
                        archivename,
                        mode="r",
                        compression=zipfile.ZIP_DEFLATED,
                        allowZip64=True,
                    )
                    namelist = c.namelist()
                    extract = [
                        e
                        for e in namelist
                        if os.path.join(self.home_directory, e) in names[n]
                    ]
                    if len(extract) != len(namelist):
                        not_backups.append(os.path.basename(archivename))
                        c.close()
                        continue
                    c.close()
                    try:
                        os.remove(archiveguard)
                    except:
                        pass
                    try:
                        os.remove(archivename)
                    except:
                        pass
                if not_backups:
                    return
            else:
                for n in names:
                    archiveguard = ".".join((n, "grd"))
                    archivename = ".".join((n, "bz2"))
                    try:
                        os.remove(archiveguard)
                    except:
                        pass
                    try:
                        os.remove(archivename)
                    except:
                        pass
        return True

    def add_import_buttons(self, *a):
        """Add button actions for Berkeley DB to Import dialogue.

        None required.  Method exists for DPT compatibility.

        """
        pass

    def get_file_sizes(self):
        """Return an empty dictionary.

        No sizes needed.  Method exists for DPT compatibility.

        """
        return dict()

    def report_plans_for_estimate(self, estimates, reporter):
        """Remind user to check estimated time to do import.

        No planning needed.  Method exists for DPT compatibility.

        """
        # See comment near end of class definition Chess in relative module
        # ..gui.chess for explanation of this change.
        # reporter.append_text_only(''.join(
        #    ('The expected duration of the import may make starting ',
        #     'it now inconvenient.',
        #     )))
        # reporter.append_text_only('')



# fonts.py
# Copyright 2009 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Make and copy chess pieces fonts for ChessTab.
"""

import tkinter
import tkinter.font

# Current practical way to determine if running on Wine, taking advantage of
# a problem found in ..core.uci which prevents UCI chess engines being used to
# do position analysis in this environment.  Here the consequence is fonts like
# 'Chess Cases' in '~/.wine/drive_c/windows/Fonts' are used as the default font
# in Font() calls, while these are not so used when put in the corresponding
# place on Microsoft Windows.
# ChessTab should be run, under Wine, by a dedicated username to avoid
# polluting the widgets of other Python programs, under Wine, with the chess
# font character set.
# The same problem occurs under Wine when these chess fonts are put in the
# /usr/local/share/fonts directory, for example, to support ChessTab on
# unix-like systems.
import multiprocessing

try:
    multiprocessing.Queue()
    _assume_wine = False
except OSError:
    _assume_wine = True
del multiprocessing

from . import constants

modify_font_attributes = ("family", "weight", "slant", "size")
modify_board_font_attributes = ("family", "weight")
integer_attributes = {"size"}


def copy_board_font_attributes(source, target):
    """Copy font attributes changeable for board from source to target."""
    _copy_font_attributes(modify_board_font_attributes, source, target)


def copy_font_attributes(source, target):
    """Copy font attributes changeable for score from source to target."""
    _copy_font_attributes(modify_font_attributes, source, target)


def make_chess_fonts(root, preferred_pieces=constants.PREFERRED_PIECES):
    """Create score and board fonts and return tuple of fonts."""
    default = _get_default_font_actual(tkinter.Text)
    moves = tkinter.font.Font(
        root=root, name=constants.MOVES_PLAYED_IN_GAME_FONT, **default
    )
    tags = tkinter.font.Font(
        root=root, name=constants.TAGS_VARIATIONS_COMMENTS_FONT, **default
    )

    # An explicit wildpieces font was not noticed to be necessary before
    # Python2.7 on Microsoft Windows.  But I had only been using 'x', for any
    # piece before and this still worked without the explicit font.
    # On FreeBSD (non-Wine) the explicit font is not needed.
    wildpieces = tkinter.font.Font(
        root=root, name=constants.WILDPIECES_ON_BOARD_FONT, **default
    )

    ff = set(tkinter.font.families())
    for pff in preferred_pieces:
        if pff in ff:
            del default["family"]
            pieces = tkinter.font.Font(
                root=root,
                name=constants.PIECES_ON_BOARD_FONT,
                family=pff,
                **default
            )
            break
    else:
        pieces = tkinter.font.Font(
            root=root, name=constants.PIECES_ON_BOARD_FONT, **default
        )
    moves.configure(weight=tkinter.font.BOLD)
    pieces.configure(weight=tkinter.font.BOLD)
    wildpieces.configure(weight=tkinter.font.BOLD)
    return moves, tags, pieces, wildpieces


def make_list_fonts(root):
    """Create game list font and return font in tuple."""
    return tkinter.font.Font(
        root=root,
        name=constants.LISTS_OF_GAMES_FONT,
        **_get_default_font_actual(tkinter.Text)
    )


def _copy_font_attributes(attributes, source, target):
    """Copy attributes from source font to target font."""
    for fa in attributes:
        if fa in source:
            target[fa] = source[fa]


def _get_default_font_actual(widget_class):
    """Return actuals for widget_class font or Courier if it does not exist."""
    if _assume_wine:
        return tkinter.font.Font(family="Courier").actual()

    # On MS Windows the named font from widget_class exists at Python2.6
    # but not at Python2.5
    try:
        return tkinter.font.nametofont(widget_class().cget("font")).actual()
    except tkinter.TclError:
        return tkinter.font.Font(family="Courier").actual()



# __init__.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Help files for ChessTab.

The options in the Help menu are named in the DATA section below.  Each is
backed by an "rst" file, displayed by invoking the menu option, and a "html"
file generated by "rst2html".

The two file names which are not obvious when viewing the directory "help" are
"keyboard", associated with ACTIONS, and "chesstab", associated with NOTES.
"""

import os

ABOUT = "About"
FILESIZE = "FileSize"
GUIDE = "Guide"
NOTES = "Notes"
SELECTION = "Selection"

_textfile = {
    ABOUT: ("aboutchesstab",),
    FILESIZE: ("filesize",),
    GUIDE: ("guide",),
    NOTES: ("chesstab",),
    SELECTION: ("selection",),
}

# Usually help files are in .../site_packages/chesstab/help but when running
# an executable generated by py2exe installed by Inno installer this module
# is in .../ChessTab/library.zip/chesstab while the help files are in
# .../ChessTab/help
folder = os.path.dirname(__file__).replace(
    os.path.join("ChessTab", "library.zip", "chesstab"), "ChessTab"
)

for k in list(_textfile.keys()):
    _textfile[k] = tuple(
        [os.path.join(folder, ".".join((n, "rst"))) for n in _textfile[k]]
    )

del folder, k, os



# displaytext.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Provide classes which define methods shared by classes in cqldisplay and
querydisplay modules which display plain text.

The cqldisplay module has two sets of classes: based on the _CQLDisplay and
CQLDialogue classes.

The querydisplay module has an identical structure where the _CQLDisplay
and _QueryDisplay classes have many methods in common.

All methods in the classes in this module existed as multiple copies in various
classes in the cqldisplay and querydisplay modules.  They are now deleted
from those modules.

The classes in this module represent the different sets of classes with methods
in common.  Two classes could reasonably be called DisplayText: the choice is
consistent with the naming of InsertText and EditText.  ShowText was chosen for
the other class.

The ShowText class is populated with the methods identical in _CQLDisplay
and _QueryDisplay, which were then removed from those two classes.

The DisplayText class is populated with the methods identical in CQLDisplay
and QueryDisplay, which were then removed from those two classes.

The InsertText class is populated with the methods identical in
CQLDisplayInsert and QueryDisplayInsert, which were then removed from those
two classes.

The EditText class is populated with the methods identical in
CQLDisplayEdit and QueryDisplayEdit, which were then removed from those two
classes.  It probably should be a subclass of InsertText, but this depends on
successful choice of method resolution order in the classes in cqldisplay and
querydisplay modules.

"""
import tkinter
import tkinter.messagebox

from solentware_grid.gui.dataedit import RecordEdit
from solentware_grid.gui.datadelete import RecordDelete

from .eventspec import EventSpec
from .score import NonTagBind


# ShowText because DisplayText fits CQLDisplay (and *Query*), but there is
# no EngineDisplay class.  ScorePGN has no equivalent in the *Text line
# because selection rules, CQL statements, and run engine commands, do not
# have 'analysis'.
class ShowText:

    """Mixin providing methods shared by the cqldisplay._CQLDisplay and
    querydisplay.QueryDisplay classes.

    Provide focus switching and widget visibility methods for widgets which
    display plain text.

    The cqldisplay.CQLDialogue and query.QueryDialogue classes are used in a
    Toplevel which displays one ChessQL or Selection Rule statement.  The
    focus switching provided here is not needed.

    """

    # The methods identical except for docstrings.  Here 'PGN score' replaces
    # 'game' and 'repertoire'.  The method names already had 'item' rather
    # than 'game' or 'repertoire'.  Perhaps 'pgn_score' is better, except
    # sometimes the method name should be compatible with the 'CQL' and
    # 'Select' classes.

    # Renamed from _bind_for_board_navigation to fit current use.
    # Before introduction of querydisplay and removal of board from predecessor
    # of cqldisplay on conversion, it was a reasonable name.
    def bind_for_item_navigation(self):
        """Set bindings to navigate text statement on pointer click."""
        self.set_score_pointer_item_navigation_bindings(True)

    def bind_for_widget_navigation(self):
        """Set bindings to give focus to this text statement on pointer click."""
        self.set_score_pointer_widget_navigation_bindings(True)

    def get_close_item_events(self):
        return ((EventSpec.display_dismiss, self.delete_item_view),)

    def set_insert_or_delete(self):
        """Convert edit display to insert display.

        PGN scores displayed for editing from a database are not closed if the
        database is closed.  They are converted to insert displays and can
        be used to add new records to databases opened later.

        """
        self.sourceobject = None

    def set_and_tag_item_text(self, reset_undo=False):
        """Delegate to superclass method and set PGN score inactive."""

        # Superclass may set self._most_recent_bindings but test below must be
        # against current value.
        mrb = self._most_recent_bindings

        super().set_and_tag_item_text(reset_undo=reset_undo)
        if mrb != NonTagBind.NO_EDITABLE_TAGS:
            for es in (self.get_inactive_button_events(),):
                self.set_event_bindings_score(es, switch=True)

    def create_database_submenu(self, menu):
        submenu = tkinter.Menu(master=menu, tearoff=False)
        self.set_popup_bindings(submenu, self.get_database_events())
        return submenu

    # The only active bindings compared with displaypgn.ShowPGN.
    def set_primary_activity_bindings(self, switch=True):
        """Delegate to toggle other relevant bindings and toggle bindings for
        database actions, navigation to other widgets, and close widget.

        """
        super().set_primary_activity_bindings(switch=switch)
        self.set_database_navigation_close_item_bindings(switch=switch)

    # Not relevant away from displaypgn.ShowPGN.
    # def set_select_variation_bindings(self, switch=True):
    #    """Delegate to toggle other relevant bindings and toggle bindings for
    #    database actions, navigation to other widgets, and close widget.

    #    """
    #    super().set_select_variation_bindings(switch=switch)
    #    self.set_database_navigation_close_item_bindings(switch=switch)

    # The methods identical except for docstrings, and references to
    # self.ui.game_items or self.ui.repertoire_items replaced by property
    # self.ui_displayed_items.

    def next_item(self, event=None):
        """Select next item on display.

        Call _next_item after 1 millisecond to allow message display.

        """
        if self.ui_displayed_items.count_items_in_stack() > 1:
            self.ui_set_find_item_games(0)
            self.score.after(1, self.try_command(self.cycle_item, self.score))

    def prior_item(self, event=None):
        """Select previous item on display.

        Call _prior_item after 1 millisecond to allow message display.

        """
        if self.ui_displayed_items.count_items_in_stack() > 1:
            self.ui_set_find_item_games(-2)
            self.score.after(
                1, self.try_command(self.cycle_item, self.score), True
            )

    # What about current_pgn_score call?
    def current_item(self, event=None):
        """Select current PGN score on display."""

        # cuiai should be referencing self given use of current_item() method,
        # but style of sibling *_item() methods is followed.
        # cuiai was cuigs in gamedisplay, and cuirs in repertoiredisplay,
        # modules originally.
        items = self.ui_displayed_items
        if items.count_items_in_stack():
            cuiai = items.active_item
            self.current_pgn_score(cuiai)
            cuiai.set_statusbar_text()

    def traverse_backward(self, event=None):
        """Give focus to previous widget type in traversal order."""
        self.set_score_pointer_widget_navigation_bindings(True)
        self.ui.give_focus_backward(self.ui_displayed_items)
        return "break"

    def traverse_forward(self, event=None):
        """Give focus to next widget type in traversal order."""
        self.set_score_pointer_widget_navigation_bindings(True)
        self.ui.give_focus_forward(self.ui_displayed_items)
        return "break"

    # The methods identical except for docstrings, and references to
    # self.ui.configure_game_grid or self.ui.configure_repertoire_grid
    # replaced by property self.ui_configure_item_list_grid.

    def cycle_item(self, prior=False):
        """Select next PGN score on display."""
        items = self.ui_displayed_items
        losefocus = items.active_item
        losefocus.bind_for_widget_navigation()
        items.cycle_active_item(prior=prior)
        self.ui_configure_item_list_grid()
        gainfocus = items.active_item
        gainfocus.set_game_list()
        gainfocus.bind_for_item_navigation()
        gainfocus.takefocus_widget.focus_set()
        gainfocus.set_statusbar_text()

    def give_focus_to_widget(self, event=None):
        """Select text item on display by mouse click."""
        self.ui.set_bindings_on_item_losing_focus_by_pointer_click()
        losefocus, gainfocus = self.ui_displayed_items.give_focus_to_widget(
            event.widget
        )
        if losefocus is not gainfocus:
            self.ui_configure_item_list_grid()
            self.score.after(
                0, func=self.try_command(self.ui_set_item_name, self.score)
            )
            self.score.after(
                0,
                func=self.try_command(gainfocus.refresh_game_list, self.score),
            )
        return "break"

    # The insert_game_database method, coerced into sameness from the methods
    # in gamedisplay._GameDisplay and repertoiredisplay._RepertoireDisplay with
    # class attibutes pgn_score_name, pgn_score_source_name, pgn_score_tags,
    # and method mark_partial_positions_to_be_recalculated, and property
    # ui_base_table.  The clarity of both common bits and differences
    # seems to justify the extra syntactic complexity.

    # Probably becomes insert_item_database() in each subclass like in
    # cqldbshow and all other *db* calls like it.

    # This was not going to be moved to displaypgn except insert_game_database,
    # which uses the class attribute pgn_score_updater, was moved.
    # Both game_updater and pgn_score_updater need more generic names.
    def game_updater(self, text):
        """Make and return a chess record containing a single PGN score."""
        updater = self.pgn_score_updater()
        updater.value.load(text)
        return updater

    # Code shared by on_game_change method in _GameDisplay and
    # _RepertoireDisplay.
    # Replace _pgn_score_ with _item_ to fit with displaypgn.ShowPGN perhaps,
    # but this one will be moved to subclasses.
    def patch_pgn_score_to_fit_record_change_and_refresh_grid(
        self, grid, instance
    ):
        if self.ui_displayed_items.is_item_panel_active(self):
            # Patch data structure to look as though the edited record has
            # been read from disk.  That means DataGrid, DisplayItems, and
            # this _GameDisplay or _RepertoireDisplay, instances.
            # self.sourceobject is DataGrid.get_visible_record(<key>), the
            # record prior to editing.
            # instance adds index structures to match those which should be
            # on database.
            # instance.newrecord is edited record including index stuff.
            key = None
            for e, k in enumerate(grid.keys):
                if instance.key.recno != k[0]:
                    key = k
                    break
            grid.close_client_cursor()
            grid.datasource.get_full_position_games(self.get_position_key())
            grid.fill_view(currentkey=key, exclude=False)


class DisplayText:

    """Mixin providing methods shared by the gamedisplay.GameDisplay
    and repertoiredisplay.RepertoireDisplay classes.

    Named DisplayPGN because ShowPGN is already taken, and when created the
    only methods in the class are delete_game_database and get_database_events.
    Both were in GameDisplay and RepertoireDisplay originally.

    insert_game_database is in ShowPGN, not here, because it is also used
    in the InsertPGN line.

    """

    def get_database_events(self):
        """Return event description tuple for PGN score database actions."""
        return (
            (EventSpec.display_insert, self.insert_item_database),
            (EventSpec.display_delete, self.delete_item_database),
        )


class InsertText:

    """Mixin providing methods shared by the gamedisplay.GameDisplayInsert
    and repertoiredisplay.RepertoireDisplayInsert classes.

    Provide methods involved in generating popup menus relevant to PGN score
    editing when inserting records.

    """

    # The methods identical except for docstrings.  Here 'PGN score' replaces
    # 'game' and 'repertoire'.  The method names already had 'item' rather
    # than 'game' or 'repertoire'.  Perhaps 'pgn_score' is better, except
    # sometimes the method name should be compatible with the 'CQL' and
    # 'Select' classes.

    def get_database_events(self):
        """Return event description tuple for PGN score database actions."""
        return ((EventSpec.display_insert, self.insert_item_database),)


# Introduced to remove create_primary_activity_popup method from InsertText
# class: it's presence prevented displaypgn.InsertPGN being a subclass of
# InsertText.
# In the PGN classes the displayed list of games depends on which token is
# current, but in the non-PGN Text classes the displayed list of games is
# determined by evaluation of a query on demand.
# When inserting or editing a record the create_primary_activity_popup method
# of this class makes an option to demand evaluation available in a popup menu.
# When showing a record the demand is implicit in making the 'show' widget
# active.
# In the PGN classes the demand is always implicit in making a token current,
# so there is no PGN equivalent of ListGamesText.
class ListGamesText:

    """Mixin providing methods shared by the cqldisplay._CQLDisplay and
    and querydisplay._QueryDisplay classes.

    Provide methods involved in generating popup menus relevant to display
    of game lists when editing records.

    """

    def create_primary_activity_popup(self):
        popup = super().create_primary_activity_popup()
        self.add_list_games_entry_to_popup(popup)
        return popup


class EditText:

    """Mixin providing methods shared by the gamedisplay.GameDisplayEdit
    and repertoiredisplay.RepertoireDisplayEdit classes.

    Provide methods involved in generating popup menus relevant to PGN score
    editing when editing records.

    """

    # The methods identical except for docstrings.  Here 'PGN score' replaces
    # 'game' and 'repertoire'.  The method names already had 'item' rather
    # than 'game' or 'repertoire'.  Perhaps 'pgn_score' is better, except
    # sometimes the method name should be compatible with the 'CQL' and
    # 'Select' classes.

    def get_database_events(self):
        """Return event description tuple for PGN score database actions."""
        return (
            (EventSpec.display_insert, self.insert_item_database),
            (EventSpec.display_update, self.update_item_database),
        )

    # update_game_database becomes update_item_database in cqldispaly and
    # querydisplay.



# fullposition.py
# Copyright 2019 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Interface to chess database for full position index providing methods which
do not need to know the particular engine used.

A superclass will include a base class for particular database engines.
"""

from ..core.filespec import POSITIONS_FIELD_DEF


class FullPosition:

    """Represent subset of games on file that match a postion."""

    def __init__(self, dbhome, dbset, dbname, newrow=None):
        """Extend to provide placeholder for position used to select games."""
        super().__init__(dbhome, dbset, dbname, newrow=newrow)

        # Position used as key to select games
        self.fullposition = None

    def get_full_position_games(self, fullposition):
        """Find game records matching full position."""
        self.fullposition = None
        if not fullposition:
            self.set_recordset(self.dbhome.recordlist_nil(self.dbset))
            return

        recordset = self.dbhome.recordlist_key(
            self.dbset,
            POSITIONS_FIELD_DEF,
            self.dbhome.encode_record_selector(fullposition),
        )

        self.set_recordset(recordset)
        self.fullposition = fullposition



# selectionds.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Interface to chess database for selection rules index.

The SelectionDS class in this module supports the DPT interface to a database.

See the ..basecore.selectionds module for the SelectionDS class for apsw,
db, and sqlite3.

"""

from solentware_grid.dpt.datasourcecursor import DataSourceCursor

from ..basecore.selection import Selection


class SelectionDS(Selection, DataSourceCursor):

    """Combine a DPT DataSourceCursor with Selection."""



# sharedtext.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Provide classes which define methods shared by classes in cqltext, score,
enginetext, and querytext modules.

The methods are originally defined in classes in those modules.  All were
identical in the CQLText and QueryText classes, and were in at most one of
the Score and EngineText classes too as identical methods.  Methods identical
in all four classes are in the blanktext.BlankText class.

The displaytext and displaypgn module do a similar thing where it matters if
the class is doing 'show', 'insert', 'delete', or 'edit', actions.  Methods
in classes in this module do not care about those differences.

"""

import tkinter

from .eventspec import EventSpec


class SharedText:

    """Provide methods shared by the cqltext.CQLText and querytext.QueryText
    classes.

    Methods with the same name are used in one or both enginetext.EngineText
    and score.Score, but the internals are different.

    """

    def set_primary_activity_bindings(self, switch=True):
        """Switch bindings for traversing query statement on or off."""
        self.set_event_bindings_score(
            self.get_F10_popup_events(
                self.post_active_menu_at_top_left, self.post_active_menu
            ),
            switch=switch,
        )
        self.set_event_bindings_score(
            self.get_button_events(buttonpress3=self.post_active_menu),
            switch=switch,
        )

    def set_score_pointer_item_navigation_bindings(self, switch):
        """Set or unset pointer bindings for game navigation."""
        self.set_event_bindings_score(
            self.get_button_events(buttonpress3=self.post_active_menu),
            switch=switch,
        )

    def get_primary_activity_events(self):
        return ()


class SharedTextEngineText:

    """Provide methods shared by the cqltext.CQLText, enginetext.EngineText,
    and querytext.QueryText, classes.

    A method with the same name may be used in score.Score, but the internals
    are different.

    """

    def get_modifier_buttonpress_suppression_events(self):
        """Return tuple of event binding definitions suppressing buttonpress
        with Control, Shift, or Alt."""
        return ()

    def post_active_menu(self, event=None):
        """Show the popup menu for chess engine definition navigation."""
        return self.post_menu(
            self.primary_activity_popup,
            self.create_primary_activity_popup,
            allowed=self.is_active_item_mapped(),
            event=event,
        )

    def post_active_menu_at_top_left(self, event=None):
        """Show the popup menu for chess engine definition navigation."""
        return self.post_menu_at_top_left(
            self.primary_activity_popup,
            self.create_primary_activity_popup,
            allowed=self.is_active_item_mapped(),
            event=event,
        )


class SharedTextScore:

    """Provide methods shared by the cqltext.CQLText, querytext.QueryText,
    and score.Score, classes.

    A method with the same name may be used in enginetext.EngineText, but the
    internals are different.

    """

    def add_cascade_menu_to_popup(
        self, label, popup, bindings=None, order=None, index=tkinter.END
    ):
        """Add cascade_menu, and bindings, to popup if not already present.

        The index is used as the label on the popup menu when visible.

        The bindings are not applied if cascade_menu is alreay in popup menu.

        """
        # Cannot see a way of asking 'Does entry exist?' other than:
        try:
            popup.index(label)
        except:
            cascade_menu = tkinter.Menu(master=popup, tearoff=False)
            popup.insert_cascade(label=label, menu=cascade_menu, index=index)
            if order is None:
                order = ()
            if bindings is None:
                bindings = {}
            for definition in order:
                function = bindings.get(definition)
                if function is not None:
                    cascade_menu.add_command(
                        label=definition[1],
                        command=self.try_command(function, cascade_menu),
                        accelerator=definition[2],
                    )

    # Subclasses which need widget navigation in their popup menus should
    # call this method.
    def create_widget_navigation_submenu_for_popup(self, popup):
        """Create and populate a submenu of popup for widget navigation.

        The commands in the submenu should switch focus to another widget.

        Subclasses should define a generate_popup_navigation_maps method and
        binding_labels iterable suitable for allowed navigation.

        """
        navigation_map, local_map = self.generate_popup_navigation_maps()
        local_map.update(navigation_map)
        self.add_cascade_menu_to_popup(
            "Navigation", popup, bindings=local_map, order=self.binding_labels
        )

    # Subclasses which need dismiss widget in a menu should call this method.
    def add_close_item_entry_to_popup(self, popup):
        """Add option to dismiss widget entry to popup.

        Subclasses must provide a delete_item_view method.

        """
        self.set_popup_bindings(popup, self.get_close_item_events())

    def create_inactive_popup(self):
        assert self.inactive_popup is None
        popup = tkinter.Menu(master=self.score, tearoff=False)
        self.set_popup_bindings(popup, self.get_inactive_events())
        self.inactive_popup = popup
        return popup

    def get_inactive_button_events(self):
        return self.get_modifier_buttonpress_suppression_events() + (
            (EventSpec.buttonpress_1, self.give_focus_to_widget),
            (EventSpec.buttonpress_3, self.post_inactive_menu),
        )

    def post_inactive_menu(self, event=None):
        """Show the popup menu for a game score in an inactive item."""
        return self.post_menu(
            self.inactive_popup, self.create_inactive_popup, event=event
        )

    def post_inactive_menu_at_top_left(self, event=None):
        """Show the popup menu for a game score in an inactive item."""
        return self.post_menu_at_top_left(
            self.inactive_popup, self.create_inactive_popup, event=event
        )

    def get_inactive_events(self):
        return (
            (EventSpec.display_make_active, self.set_focus_panel_item_command),
            (EventSpec.display_dismiss_inactive, self.delete_item_view),
        )



# apswdu_dir.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import directory of PGN files with apsw.chessapswdu to database."""


if __name__ == "__main__":

    from .directory_widget import DirectoryWidget
    from ..apsw.chessapswdu import chess_sqlite3du

    DirectoryWidget(chess_sqlite3du, "apsw")



# analysisds.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Interface to chess database for chess engine analysis index.

The AnalysisDS class in this module supports the DPT interface to a database.

See the ..basecore.analysisds module for the AnalysisDS class for apsw, db,
and sqlite3.

"""

from solentware_grid.dpt.datasourcecursor import DataSourceCursor

from ..basecore.analysis import Analysis


class AnalysisDS(Analysis, DataSourceCursor):

    """Combine a DPT DataSourceCursor with Analysis."""



# positionscore.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""A chess game score display class highlighting moves for a position.

List of classes

PositionScore

"""

import tkinter

from solentware_misc.gui.exceptionhandler import ExceptionHandler

from pgn_read.core.constants import (
    SEVEN_TAG_ROSTER,
    FEN_WHITE_ACTIVE,
    DEFAULT_TAG_RESULT_VALUE,
    DEFAULT_TAG_VALUE,
    TAG_WHITE,
    TAG_BLACK,
    TAG_RESULT,
)
from pgn_read.core.parser import PGN
from pgn_read.core.squares import Squares

from ..core.pgn import GameDisplayMoves
from .constants import (
    LINE_COLOR,
    MOVE_COLOR,
    ALTERNATIVE_MOVE_COLOR,
    VARIATION_COLOR,
    MOVES_PLAYED_IN_GAME_FONT,
    TAGS_VARIATIONS_COMMENTS_FONT,
    NAVIGATE_TOKEN,
    NAVIGATE_MOVE,
    TOKEN,
    RAV_MOVES,
    CHOICE,
    PRIOR_MOVE,
    RAV_SEP,
    RAV_TAG,
    ALL_CHOICES,
    POSITION,
    MOVE_TAG,
    SELECTION,
    ALTERNATIVE_MOVE_TAG,
    LINE_TAG,
    VARIATION_TAG,
    LINE_END_TAG,
    START_SCORE_MARK,
    NAVIGATE_COMMENT,
    TOKEN_MARK,
    PGN_TAG,
    ALTERNATIVE_MOVE_COLOR,
    VARIATION_COLOR,
    SPACE_SEP,
    NEWLINE_SEP,
    NULL_SEP,
    MOVETEXT_MOVENUMBER_TAG,
)


# May need to make this a superclass of Tkinter.Text because DataRow method
# make_row_widgets expects to be able to call Tkinter widget methods.
class PositionScore(ExceptionHandler):

    """Chess game score widget composed from a Text widget."""

    m_color = MOVE_COLOR
    am_color = ALTERNATIVE_MOVE_COLOR
    v_color = VARIATION_COLOR
    tags_variations_comments_font = TAGS_VARIATIONS_COMMENTS_FONT
    moves_played_in_game_font = MOVES_PLAYED_IN_GAME_FONT

    tags_displayed_last = SEVEN_TAG_ROSTER

    def __init__(
        self,
        widget,
        tags_variations_comments_font=None,
        moves_played_in_game_font=None,
        ui=None,
        **ka
    ):
        """Extend with widgets to display game.

        widget - Tkinter.Text instance to contain game score
        tags_variations_comments_font - font for tags variations and comments
        moves_played_in_game_font - font for move played in game
        ui - the ChessUI instance

        Create Frame in toplevel and add Canvas and Text.
        Text width and height set to zero so widget fit itself into whatever
        space Frame has available.
        Canvas must be square leaving Text at least half the Frame.

        """
        super(PositionScore, self).__init__()
        self.ui = ui
        if tags_variations_comments_font:
            self.tags_variations_comments_font = tags_variations_comments_font
        if moves_played_in_game_font:
            self.moves_played_in_game_font = moves_played_in_game_font
        # Use widget argument rather than create one here like similar classes.
        self.score = widget
        widget.tag_configure(MOVETEXT_MOVENUMBER_TAG, elide=tkinter.FALSE)
        widget.tag_configure(
            MOVES_PLAYED_IN_GAME_FONT, font=self.moves_played_in_game_font
        )
        # Order is MOVE_TAG ALTERNATIVE_MOVE_TAG VARIATION_TAG so that correct
        # colour has highest priority as moves are added to and removed from
        # tags.
        widget.tag_configure(MOVE_TAG, background=self.m_color)
        widget.tag_configure(ALTERNATIVE_MOVE_TAG, background=self.am_color)
        widget.tag_configure(VARIATION_TAG, background=self.v_color)
        widget.bind("<Map>", self.try_event(self.on_map))
        # None implies initial position and is deliberately not a valid Tk tag.
        self.current = None  # Tk tag of current move
        self._clear_tag_maps()
        self.collected_game = None
        # map_game uses self._force_movenumber to insert movenumber before
        # black moves after start and end of recursive annotation variation
        # markers.
        self._force_movenumber = False

    # Nothing needed on <Unmap> event at present
    def on_map(self, event):
        """Scroll text of move into current position to left edge of widget.

        This is done correctly only if the widget is mapped.  So do it when
        a Map event occurs.

        """
        pass

    def _clear_tag_maps(self):
        """Clear mappings of tags and positions.

        Instances of PositionScore are reused as games are navigated and not
        necessarely for the game already in the instance.

        """
        # Mappings tagpositionmap, fullpositions, positiontags, and recently
        # added varmovetags, are removed because the compare_two_positions()
        # method and _context attributes seems to be able to do the job.
        # Probably including taking account of castling and en-passant options
        # if required.
        self.variation_number = 0
        self.varstack = []
        self.position_number = 0

    def process_score(self, text=None, context=None):
        """Wrapper for Tkinter.Text configure method for score attribute"""
        if text:
            self._clear_tag_maps()
            self.collected_game = next(
                PGN(game_class=GameDisplayMoves).read_games(text)
            )
            self._context = context
            try:
                self.set_and_tag_item_text()
            finally:
                del self._context

    def get_top_widget(self):
        """Return topmost widget for game display.

        The topmost widget is put in a container widget in some way
        """
        return self.score

    def destroy_widget(self):
        """Destroy the widget displaying game."""
        self.score.destroy()

    def set_and_tag_item_text(self, reset_undo=False):
        """Display the game as board and moves.

        reset_undo causes the undo redo stack to be cleared if True.  Set True
        on first display of a game for editing so that repeated Ctrl-Z in text
        editing mode recovers the original score.

        """
        self.score.configure(state=tkinter.NORMAL)
        self.score.delete("1.0", tkinter.END)
        self.map_game()
        self.score.configure(state=tkinter.DISABLED)

    def clear_current_range(self):
        """Remove existing MOVE_TAG ranges."""
        tr = self.score.tag_ranges(MOVE_TAG)
        if tr:
            self.score.tag_remove(MOVE_TAG, tr[0], tr[1])

    def get_tags_display_order(self):
        """Return tags in alphabetic order modified by self.tags_displayed_last.

        last=None means do not display the tags: assume tags will be displayed
        in the order they appear in the PGN text.

        Return None if last is None, and list of '(tag, value)'s otherwise.

        last modifies the order PGN tags are displayed.  Normally the Seven
        Tag Roster appears first in a PGN game score followed by other tags
        in alphabetic order.  Tags not in last are displayed in alphabetic
        order followed by the tags in last.  If last is None the PGN tags are
        displayed in the order they appear in the PGN game score.

        The intention is to display the important tags adjacent to the game
        score.  Thus if last is the Seven Tag Roster these tags are displayed
        after the other tags, rather than appearing before the other tags as
        in a PGN file.
        """
        last = self.tags_displayed_last
        if last is None:
            return None
        tag_values = []
        tags = self.collected_game._tags
        for tv in sorted(tags.items()):
            if tv[0] not in last:
                tag_values.append(tv)
        for t in last:
            if t in tags:
                tag_values.append((t, tags[t]))
        return tag_values

    def get_variation_tag_names(self):
        """Return suffixed RAV_MOVES tag names.

        The suffixes are arbitrary so increment then generate suffix would be
        just as acceptable but generate then increment uses all numbers
        starting at 0.

        """
        self.variation_number += 1
        suffix = str(self.variation_number)
        return "".join((RAV_MOVES, suffix))

    def get_next_positiontag_name(self):
        """Return suffixed POSITION tag name."""
        self.position_number += 1
        return "".join((POSITION, str(self.position_number)))

    def get_position_tag_of_index(self, index):
        """Return Tk tag name if index is in a position tag"""
        for tn in self.score.tag_names(index):
            if tn.startswith(POSITION):
                return tn
        return None

    def get_current_tag_and_mark_names(self):
        """Return suffixed POSITION and TOKEN tag and TOKEN_MARK mark names."""
        suffix = str(self.position_number)
        return ["".join((t, suffix)) for t in (POSITION, TOKEN, TOKEN_MARK)]

    def get_tag_and_mark_names(self):
        """Return suffixed POSITION and TOKEN tag and TOKEN_MARK mark names.

        The suffixes are arbitrary so increment then generate suffix would be
        just as acceptable but generate then increment uses all numbers
        starting at 0.

        A TOKEN_MARK name is generated for each token but the mark will be
        created only for editable tokens.

        """
        self.position_number += 1
        suffix = str(self.position_number)
        return ["".join((t, suffix)) for t in (POSITION, TOKEN, TOKEN_MARK)]

    def insert_token_into_text(self, token, separator):
        """Insert token and separator in widget.  Return boundary indicies.

        Indicies for start and end of token text are noted primarily to control
        editing and highlight significant text.  The end of separator index is
        used to generate contiguous regions for related tokens and act as a
        placeholder when there is no text between start and end.

        """
        widget = self.score
        start = widget.index(tkinter.INSERT)
        widget.insert(tkinter.INSERT, token)
        end = widget.index(tkinter.INSERT)
        widget.insert(tkinter.INSERT, separator)
        return start, end, widget.index(tkinter.INSERT)

    def is_currentmove_in_main_line(self):
        """Return True if currentmove is in the main line tag"""
        return self.is_index_in_main_line(
            self.score.tag_ranges(self.current)[0]
        )

    def is_index_in_main_line(self, index):
        """Return True if index is in the main line tag"""
        return bool(
            self.score.tag_nextrange(
                self.gamevartag, index, "".join((str(index), "+1 chars"))
            )
        )

    def is_move_in_main_line(self, move):
        """Return True if move is in the main line"""
        return self.is_index_in_main_line(self.score.tag_ranges(move)[0])

    """At present both POSITION<suffix> and TOKEN<suffix> tags exist.

    Now that setting MOVE_TAG has been moved to token-type specific code there
    is probably no need for both.  TOKEN marks the active text of a POSITION,
    for example {<active text>} where POSITION includes the surrounding braces,
    but as active text can vary and when null POSITION is used to set MOVE_TAG
    might as well refer to POSITION directly.

    This means the set_insert_mark_for_editing and set_bound_marks_for_editing
    methods do not need to loop through tag names for TOKEN because it starts
    from a POSITION tag name.

    """

    def _set_square_piece_map(self, position):
        assert len(position) == 1
        spm = self._square_piece_map
        spm.clear()
        for p, s in position[0][0]:
            spm[s] = p

    def _modify_square_piece_map(self, position):
        assert len(position) == 2
        spm = self._square_piece_map
        for s, p in position[0][0]:
            del spm[s]
        for s, p in position[1][0]:
            spm[s] = p

    # Attempt to re-design map_game method to fit new pgn_read package.
    def map_game(self):
        """Tag and mark the displayed text of game score.

        The tags and marks are used for colouring and navigating the score.

        """
        self._force_movenumber = True

        # With get_current_...() methods as well do not need self._vartag
        # state attributes.
        self._vartag = self.get_variation_tag_names()
        self.gamevartag = self._vartag

        self._square_piece_map = {}

        cg = self.collected_game
        spm = self._square_piece_map
        for p, s in cg._initial_position[0]:
            spm[s] = p
        assert len(cg._text) == len(cg._position_deltas)
        for text, delta in zip(cg._text, cg._position_deltas):
            t0 = text[0]
            if t0 in "abcdefghKQRBNkqrnO":
                self.map_move_text(text, delta)
            elif t0 == "(":
                self.map_start_rav(text, delta)
            elif t0 == ")":
                self.map_end_rav(text, delta)
            elif t0 in "10*":
                self.map_termination(text)
            else:
                self.map_non_move(text)
        self.insert_token_into_text(
            cg._tags.get(TAG_WHITE, DEFAULT_TAG_VALUE), SPACE_SEP
        )
        self.insert_token_into_text(
            cg._tags.get(TAG_RESULT, DEFAULT_TAG_RESULT_VALUE), SPACE_SEP
        )
        self.insert_token_into_text(
            cg._tags.get(TAG_BLACK, DEFAULT_TAG_VALUE), SPACE_SEP
        )

        tr = self.score.tag_nextrange(NAVIGATE_MOVE, "1.0")
        if tr:
            self.score.mark_set(START_SCORE_MARK, str(tr[0]))
        else:
            self.score.mark_set(START_SCORE_MARK, "1.0")

        # Delete the attributes used to build the self.score Text widget.
        del self._square_piece_map
        del self._force_movenumber
        del self._vartag

    def compare_two_positions(self, one, two):
        """Return True if positions one and two are same, otherwise False.

        Ignore castling and en passant options when comparing positions.

        The move on which the positions occur is not relevant so full move
        number and half move clock are ignored.

        """
        # At end of game or variation either one or two will be None, depending
        # on argument order.
        # Exception return must be False because True causes too much display.
        try:
            if one[1] != two[1]:
                return False
        except TypeError:
            if one is None or two is None:
                return False
            raise
        if set(one[0]) != set(two[0]):
            return False
        two0 = two[0]
        for s, p in one[0].items():
            if p.name != two0[s].name:
                return False
        return True

    def map_move_text(self, token, delta):
        """Add token to game text and modify game position by delta"""
        prevcontext, currcontext, nextcontext = self._context

        # Does position in which move is played match the previous or current
        # position in self._context?
        prev_position = (self._square_piece_map.copy(),) + delta[0][1:]
        prev_match_prevcontext = self.compare_two_positions(
            prev_position, prevcontext
        )
        prev_match_currcontext = self.compare_two_positions(
            prev_position, currcontext
        )

        self._modify_square_piece_map(delta)

        # Does position after move is played match the current or next
        # position in self._context?
        curr_position = (self._square_piece_map.copy(),) + delta[1][1:]
        curr_match_currcontext = self.compare_two_positions(
            curr_position, currcontext
        )
        next_match_nextcontext = self.compare_two_positions(
            curr_position, nextcontext
        )

        if not (
            prev_match_prevcontext
            or prev_match_currcontext
            or curr_match_currcontext
            or next_match_nextcontext
        ):
            return
        widget = self.score
        if delta[1][1] != FEN_WHITE_ACTIVE:
            start, end, sepend = self.insert_token_into_text(
                str(delta[1][5]) + ".", SPACE_SEP
            )
            widget.tag_add(MOVETEXT_MOVENUMBER_TAG, start, sepend)
        elif self._force_movenumber:
            start, end, sepend = self.insert_token_into_text(
                str(delta[0][5]) + "...", SPACE_SEP
            )
            widget.tag_add(MOVETEXT_MOVENUMBER_TAG, start, sepend)
        self._force_movenumber = False
        positiontag = self.get_next_positiontag_name()
        start, end, sepend = self.insert_token_into_text(token, SPACE_SEP)
        for tag in positiontag, self._vartag, NAVIGATE_MOVE:
            widget.tag_add(tag, start, end)
        if self._vartag is self.gamevartag:
            widget.tag_add(MOVES_PLAYED_IN_GAME_FONT, start, end)
        widget.tag_add("".join((RAV_SEP, self._vartag)), start, sepend)
        if not (prev_match_prevcontext or prev_match_currcontext):
            # token is move to reach position and different from context
            widget.tag_add(VARIATION_TAG, start, end)
            pass
        if not (curr_match_currcontext or next_match_nextcontext):
            # token is move out of position and  different from context
            widget.tag_add(ALTERNATIVE_MOVE_TAG, start, end)
            pass
        tr = widget.tag_prevrange(self._vartag, start)
        if not tr:
            varstack = list(self.varstack)
            while varstack:
                var = varstack.pop()
                tpr = widget.tag_prevrange(var, start)
                if not tpr:
                    break
                tr = widget.tag_prevrange(var, tpr[0])
                if tr:
                    break

    def map_start_rav(self, token, position):
        """Add token to game text. position is ignored. Return range and prior.

        Variation tags are set for guiding move navigation. self._vartag
        is placed on
        a stack for restoration at the end of the variation.

        """
        self._set_square_piece_map(position)
        widget = self.score
        self.varstack.append(self._vartag)
        self._vartag = self.get_variation_tag_names()
        self.insert_token_into_text(token, SPACE_SEP)
        self._force_movenumber = True

    def map_end_rav(self, token, position):
        """Add token to game text. position is ignored. Return token range.

        Variation tags are set for guiding move navigation. self._vartag
        is restored
        from the stack for restoration at the end of the variation.

        """
        self._set_square_piece_map(position)
        start, end, sepend = self.insert_token_into_text(token, SPACE_SEP)
        self._vartag = self.varstack.pop()
        self._force_movenumber = True

    def map_termination(self, token):
        """Add token to game text. position is ignored. Return token range."""
        if self.collected_game._tags.get(TAG_RESULT) != token:
            self.insert_token_into_text(token, SPACE_SEP)

    def map_non_move(self, token):
        """Add token to game text. position is ignored. Return token range."""
        # self.insert_token_into_text(token, SPACE_SEP)

    def set_current(self):
        """Remove existing MOVE_TAG ranges and add self.currentmove ranges.

        Subclasses may adjust the MOVE_TAG range if the required colouring
        range of the item is different.  For example just <text> in {<text>}
        which is a PGN comment where <text> may be null after editing.

        The adjusted range must be a subset of self.currentmove range.

        """
        # Superclass set_current method may adjust bindings so do not call
        # context independent binding setup methods after this method for
        # an event.
        tr = self.set_current_range()
        if tr:
            self.set_move_tag(tr[0], tr[1])
            return tr

    def set_current_range(self):
        """Remove existing MOVE_TAG ranges and add self.currentmove ranges.

        Subclasses may adjust the MOVE_TAG range if the required colouring
        range of the item is different.  For example just <text> in {<text>}
        which is a PGN comment where <text> may be null after editing.

        The adjusted range must be a subset of self.currentmove range.

        """
        self.clear_current_range()
        if self.current is None:
            return
        tr = self.score.tag_ranges(self.current)
        if not tr:
            return
        return tr

    def set_move_tag(self, start, end):
        """Add range start to end to MOVE_TAG (which is expected to be empty).

        Assumption is that set_current_range has been called and MOVE_TAG is
        still empty following that call.

        """
        self.score.tag_add(MOVE_TAG, start, end)



# gamedbedit.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Customise edit toplevel to edit or insert chess game record.
"""

import tkinter
import tkinter.messagebox

from solentware_grid.gui.dataedit import DataEdit

from solentware_misc.gui.exceptionhandler import ExceptionHandler

from pgn_read.core.parser import PGN
from pgn_read.core.constants import TAG_WHITE, TAG_BLACK

from .gametoplevel import GameToplevel, GameToplevelEdit
from .toplevelpgn import EditPGN
from .constants import EMPTY_SEVEN_TAG_ROSTER


class GameDbEdit(ExceptionHandler, EditPGN, DataEdit):

    """Edit PGN text for game on database, or insert a new record.

    parent is used as the master argument in GameToplevel calls.

    ui is used as the ui argument in GameToplevel calls.

    newobject, parent, oldobject, and the one or two GameToplevel instances
    created, are used as arguments in the super.__init__ call.

    showinitial determines whether a GameToplevel is created for oldobject if
    there is one.

    Attribute pgn_score_name provides the name used in widget titles and
    message text.

    Attribute pgn_score_tags provides empty PGN tags to present when creating
    an insert Toplevel.  It is the empty PGN Seven Tag Roster.

    Attribute pgn_score_source provides the error key value to index a PGN
    game score with errors.

    Methods get_title_for_object and set_item, and properties ui_base_table;
    ui_items_in_toplevels; and ui, allow similar methods in various classes
    to be expressed identically and defined once.

    """

    pgn_score_name = "Game"
    pgn_score_tags = EMPTY_SEVEN_TAG_ROSTER
    pgn_score_source = "Editor"

    def __init__(
        self, newobject, parent, oldobject, showinitial=True, ui=None
    ):
        """Extend and create dialogue widget to edit or insert chess game."""
        if not oldobject:
            showinitial = False
        super().__init__(
            newobject,
            parent,
            oldobject,
            GameToplevelEdit(master=parent, ui=ui),
            "",
            oldview=GameToplevel(master=parent, ui=ui)
            if showinitial
            else showinitial,
        )
        self.initialize()

    @property
    def ui_base_table(self):
        return self.ui.base_games

    @property
    def ui_items_in_toplevels(self):
        return self.ui.games_and_repertoires_in_toplevels

    @property
    def ui(self):
        return self.newview.ui

    def set_item(self, view, object_):
        self.set_default_source_for_object(object_)
        view.set_position_analysis_data_source()
        view.collected_game = next(
            PGN(game_class=view.gameclass).read_games(object_.get_srvalue())
        )
        view.set_and_tag_item_text()

    def get_title_for_object(self, object_=None):
        """Return title for Toplevel containing a Game object_.

        Default value of object_ is oldobject attribute from DataEdit class.

        """
        if object_ is None:
            object_ = self.oldobject
        if object_:
            tags = object_.value.collected_game._tags
            try:
                return "  ".join(
                    (
                        self.pgn_score_name.join(("Edit ", ":")),
                        " - ".join((tags[TAG_WHITE], tags[TAG_BLACK])),
                    )
                )
            except TypeError:
                return self.pgn_score_name.join(
                    ("Edit ", " - names unknown or invalid")
                )
            except KeyError:
                return self.pgn_score_name.join(
                    ("Edit ", " - names unknown or invalid")
                )
        else:
            return "".join(("Insert ", self.pgn_score_name))

    def set_default_source_for_object(self, object_=None):
        """Set default source for Toplevel containing a Game object_.

        Default value of object_ is oldobject attribute from DataEdit class.

        Currently do nothing for games.  Originally used for games with PGN
        errors, where it was the name of the PGN file containing the game.

        Now present for compatibility with Repertoires.

        """
        pass

    def put(self):
        """Mark partial position records for recalculation and return key."""
        self.datasource.dbhome.mark_partial_positions_to_be_recalculated()
        super().put()

    def edit(self):
        """Mark partial position records for recalculation and return key."""
        self.datasource.dbhome.mark_partial_positions_to_be_recalculated()
        super().edit()
